//! Handles inserting DRS file metadata into Solr

use std::io;
use std::sync::Arc;
use std::time::Duration;
use std::{collections::HashMap, path::Path};

use camino::Utf8PathBuf;
use thiserror::Error;
use tokio::sync::mpsc::{channel, Receiver};
use tokio::time::Instant;
use tracing::{error, info, instrument, trace, warn};
use walkdir::{DirEntry, WalkDir};

use crate::drs::Structure;
use crate::{
    drs::{Config, Metadata},
    solr::{Solr, SolrError},
};
use crate::{DEFAULT_FILES_COLLECTION, DEFAULT_LATEST_COLLECTION};

/// Errors possible while ingesting data into solr
#[derive(Debug, Error)]
pub enum IngestError {
    /// Error communicating with or inserting data into Solr
    #[error("error inserting into store")]
    InsertFail(#[from] SolrError),

    /// Error crawling the filesystem tree. Most likely a symlink loop or a broken file. See underlying error for more
    /// details.
    #[error("error while walking the directory structure")]
    WalkDirectory(#[from] walkdir::Error),

    /// Error canonicalizing data dir path
    #[error("error canonicalizing data-dir, may not exist")]
    DataDirCanonicalization(#[from] io::Error),
}

#[derive(Debug)]
pub struct IngestReport {
    pub structures: Vec<StructureReport>,
    pub total_time: Duration,
}

#[derive(Debug)]
pub struct StructureReport {
    pub structure: Structure,
    pub sent: u32,
    pub skipped: u32,
    pub time: Duration,
}

/// Ingests metadata from filesystem into freva's solr
///
/// Spawns 2 tasks per [`Structure`] in the DRS config where `root_dir` is contained in (i.e. `starts_with`) `data_dir`.
/// One task is the computation thread which processes the file names as they come in and it spawns the second task
/// which is a blocking thread that uses `walkdir` synchronously and feeds the paths it finds into a channel.
///
/// `files_collection` and `latest_collection` are optional. If `None`, then this will use Freva's defaults:
/// [`DEFAULT_FILES_COLLECTION`] and [`DEFAULT_LATEST_COLLECTION`], respectively.
pub async fn ingest<P: AsRef<Path>>(
    solr: &Solr,
    drs_conf: &Config,
    data_dir: P,
    batch_size: usize,
    files_collection: Option<&str>,
    latest_collection: Option<&str>,
) -> Result<IngestReport, IngestError> {
    let mut handles = Vec::new();
    let data_dir = tokio::fs::canonicalize(data_dir.as_ref()).await?;

    let start_time = Instant::now();

    for structure in drs_conf.structures() {
        if structure.root().starts_with(&data_dir) {
            // clone for each thread to capture without worrying about lifetimes
            let solr = solr.clone();
            let structure = structure.clone();
            let files_collection = files_collection
                .unwrap_or(DEFAULT_FILES_COLLECTION)
                .to_owned();
            let latest_collection = latest_collection
                .unwrap_or(DEFAULT_LATEST_COLLECTION)
                .to_owned();
            handles.push(tokio::spawn(async move {
                ingest_structure(
                    solr,
                    structure,
                    batch_size,
                    &files_collection,
                    &latest_collection,
                )
                .await
            }));
        }
    }

    let results = futures::future::join_all(handles).await;

    let (success, errors): (Vec<_>, Vec<_>) = results.into_iter().partition(|r| r.is_ok());
    for err in errors.into_iter().map(|e| e.unwrap_err()) {
        error!("error while processing a DRS structure: {}", err);
    }

    let (results, errors): (Vec<_>, Vec<_>) = success
        .into_iter()
        .map(|s| s.unwrap())
        .partition(|r| r.is_ok());

    for err in errors.into_iter().map(|e| e.unwrap_err()) {
        error!("error while processing a DRS structure: {}", err);
    }

    let reports = results.into_iter().map(|r| r.unwrap()).collect();

    Ok(IngestReport {
        structures: reports,
        total_time: start_time.elapsed(),
    })
}

/// Ingests a single structure from the DRS config.
///
/// This handles crawling the structure's `root_dir` and ingesting any files it finds there are valid for the structure.
/// This inserts the files it finds into both files and latest collection within solr.
///
/// Invalid files, whatever the reason they're invalid, will be skipped. Failing to send to solr will result in this
/// ending early with an error. Any files that were already sent to solr will not be cleaned up.
#[instrument(skip(solr, structure), fields(structure = structure.activity().as_str()))]
async fn ingest_structure(
    solr: Solr,
    structure: Structure,
    batch_size: usize,
    files_collection: &str,
    latest_collection: &str,
) -> Result<StructureReport, IngestError> {
    let mut file_buf = Vec::with_capacity(batch_size);
    let mut latest_buf = Vec::with_capacity(batch_size);

    let mut sent = 0;
    let mut skipped = 0;
    let mut latest_versions: HashMap<String, String> = HashMap::new();

    let walkdir = WalkDir::new(structure.root()).follow_links(true);
    let mut entries = walkdir_channel(walkdir);

    let start_time = Instant::now();

    while let Some(entry) = entries.recv().await {
        if file_buf.len() == batch_size {
            sent += flush(&solr, files_collection, &mut file_buf).await?;
        }
        if latest_buf.len() == batch_size {
            flush(&solr, latest_collection, &mut latest_buf).await?;
        }

        let entry = match entry {
            Ok(entry) => entry,
            Err(e) => {
                // if the file isn't found, something is probably broken with Lustre, skip it and move on.
                // other errors will still result in ingestion ending early
                let path = e.path().unwrap_or_else(|| Path::new("")).display();
                if let Some(inner) = e.io_error() {
                    if inner.kind() == io::ErrorKind::NotFound {
                        warn!("found broken file, skipping: {path}");
                        skipped += 1;
                        continue;
                    }
                }
                return Err(IngestError::WalkDirectory(e));
            }
        };

        if entry.path().is_file() {
            let path = match Utf8PathBuf::from_path_buf(entry.path().to_path_buf()) {
                Ok(p) => p,
                Err(e) => {
                    warn!(
                        "not a valid drs file, has non UTF8 characters in path:\n{:?}",
                        e
                    );
                    skipped += 1;
                    continue;
                }
            };

            let f = match structure.metadata_from_path(&path) {
                Ok(f) => f,
                Err(e) => {
                    warn!("not a valid drs file, skipping:\n{:#?}", e);
                    skipped += 1;
                    continue;
                }
            };
            trace!("found document to send to solr: {}", f.to_path());

            let f = Arc::new(f);
            if let Some(version) = f.version() {
                let id = f.to_unversioned_identifier();

                match latest_versions.get_mut(&id) {
                    Some(latest) => {
                        if version >= latest.as_str() {
                            *latest = version.to_owned();
                            latest_buf.push(f.clone());
                        }
                    }
                    None => {
                        latest_versions.insert(id, version.to_owned());
                        latest_buf.push(f.clone());
                    }
                }
            } else {
                // not versioned, always add to latest
                latest_buf.push(f.clone());
            }

            file_buf.push(f.clone());
        }
    }

    if !file_buf.is_empty() {
        sent += flush(&solr, files_collection, &mut file_buf).await?;
    }
    if !latest_buf.is_empty() {
        flush(&solr, latest_collection, &mut latest_buf).await?;
    }

    Ok(StructureReport {
        structure,
        sent,
        skipped,
        time: start_time.elapsed(),
    })
}

async fn flush(
    solr: &Solr,
    collection: &str,
    docs: &mut Vec<Arc<Metadata<'_>>>,
) -> Result<u32, IngestError> {
    info!(
        "flushing {} documents in buffer to collection {}",
        docs.len(),
        collection
    );
    let sent = docs.len() as u32;
    // turn Vec<Rc<_>> in Vec<&_>
    let doc_refs: Vec<_> = docs.iter().map(|d| d.as_ref()).collect();
    solr.update(collection, &doc_refs).await?;
    docs.clear();
    Ok(sent)
}

/// Creates a channel for offloading filesystem access onto a separate task.
///
/// This is necessary because the filesystem can be extremely slow when dealing with large numbers of files but walkdir
/// doesn't provide an async API that would be more appropriate. Instead we throw it into a separate task which goes
/// into tokio's blocking pool and should result in less contention.
fn walkdir_channel(walkdir: WalkDir) -> Receiver<walkdir::Result<DirEntry>> {
    let (tx, rx) = channel(100);
    tokio::task::spawn_blocking(move || {
        for entry in walkdir {
            if let Err(e) = tx.blocking_send(entry) {
                error!("{}", e);
                return;
            }
        }
    });
    rx
}
