//! Implementation of the DRS specification for extracting metadata information from file paths.

use std::{
    collections::HashMap,
    fmt::{Display, Formatter},
    fs::File,
    io::{self, Read},
    path::{Path, PathBuf},
};

use camino::Utf8PathBuf;
use serde::{ser::SerializeMap, Deserialize, Serialize};
use thiserror::Error;
use tracing::{debug, error, warn};

/// Errors that can be returned by DRS functions
#[derive(Debug, Error)]
pub enum Error {
    /// Error while interfacing with the DRS config file
    #[error("error opening drs config file: {0}")]
    OpenConfig(#[from] io::Error),

    /// Error canonicalizing paths.
    /// This probably means that the path doesn't exist, which `canonicalize` requires.
    #[error("error canonicalizing path: {path:?}")]
    CanonicalizePath {
        /// path that failed to be canonicalized
        path: PathBuf,
    },

    /// Error parsing DRS config file
    #[error("error parsing drs config file: {0}")]
    ParseConfig(#[from] toml::de::Error),

    /// Error where path is not valid for a given [`Structure`]
    #[error("{path:?} is not a valid DRS file: {reason}")]
    InvalidPath {
        /// path that is not valid
        path: PathBuf,
        /// reason it was considered invalid
        reason: String,
    },

    /// Error that happens when given a path that isn't valid utf 8
    #[error("path is not utf-8")]
    NonUtf8Path(#[from] camino::FromPathBufError),
}

/// newtype pattern around DRS activity string
#[derive(Debug, Deserialize, PartialEq, Eq, PartialOrd, Ord, Hash, Clone)]
pub struct Activity(pub String);

impl Activity {
    /// Gets the Activity as a string. Convenience method for accessing its value as a string without using tuple
    /// syntax.
    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl Display for Activity {
    fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), std::fmt::Error> {
        write!(f, "{}", self.0)
    }
}

/// Metadata for DRS metadata derived from a file's path on the filesystem.
#[derive(Debug, Clone, Deserialize, PartialEq)]
pub struct Structure {
    /// The base directory where all DRS files of this structure will be found.
    root_dir: Utf8PathBuf,
    /// The keys of the different path segments in the order they are expected to appear in the path.
    parts_dir: Vec<String>,
    /// Key ordering for creating the dataset identifier (See 3.4 of DRS spec)
    parts_dataset: Vec<String>,
    /// Key ordering for creating the versioned dataset identifier
    parts_versioned_dataset: Option<Vec<String>>,
    /// Key ordering from parts of the file (separated by `_`)
    parts_file_name: Vec<String>,
    /// Formatting of the time section of the file's name.
    ///
    /// Example: `start_time-end_time`
    parts_time: String,
    /// Default values used in constructing the dataset identifier if not present
    defaults: HashMap<String, Utf8PathBuf>,
    /// Activity name of this DRS structure.
    ///
    /// *Note*: this does not always match the key in the [`Config`] map.
    data_type: Activity,
}

impl Structure {
    /// Performs a basic check of if the given path could possibly be valid for this structure.
    ///
    /// This is not comprehensive in that there are still paths where this would return `true` but which are not valid
    /// for this structure. For example, if a path has the same `root_dir` and the same _number_ of components in its
    /// directory structure, it will pass this even if it doesn't have the right `filename` structure for this.
    /// TODO: fix this if possible.
    pub fn verify_path<P: AsRef<Path>>(&self, path: P) -> bool {
        let path = path.as_ref();

        // Prefix must match. If not, then this isn't a valid structure for this path.
        let path = match path.strip_prefix(&self.root_dir) {
            Ok(p) => p,
            Err(_) => return false,
        };

        let parts_len = path.iter().count();
        parts_len == self.parts_dir.len()
    }

    /// Gets the root directory for this structure
    pub fn root(&self) -> &Utf8PathBuf {
        &self.root_dir
    }

    /// Gets the type of activity of this structure.
    pub fn activity(&self) -> &Activity {
        &self.data_type
    }

    /// Extracts [`Metadata`] for a file path of this structure
    pub fn metadata_from_path<P: AsRef<Path>>(&self, path: P) -> Result<Metadata, Error> {
        let path = Utf8PathBuf::try_from(path.as_ref().to_owned())?;

        if !self.verify_path(&path) {
            return Err(Error::InvalidPath {
                path: path.into(),
                reason: "file path does not match expected structure".to_owned(),
            });
        }

        let parts = path
            .strip_prefix(self.root())
            .unwrap() // this is already done with `verify_path` so should be safe
            .iter();

        let mut path_parts = HashMap::new();

        for (part, name) in parts.zip(self.parts_dir.iter()) {
            path_parts.insert(name.clone(), Utf8PathBuf::from(part));
        }

        // split file name
        // remove file extension before splitting
        // if no extension, use the whole name
        let name_no_extension = match path_parts["file_name"].file_stem() {
            Some(n) => n,
            None => path_parts["file_name"].as_str(),
        };
        let mut file_name_parts: Vec<_> =
            name_no_extension.split('_').map(|s| s.to_owned()).collect();

        // TODO: I'm not sure why this is here or what it's trying to accomplish
        if file_name_parts.len() == self.parts_file_name.len() - 1
            && file_name_parts.contains(&"fx".to_owned())
        {
            // original pushed None which doesn't work here
            file_name_parts.push("".to_owned());
        }

        for (i, name) in self.parts_file_name.iter().enumerate() {
            if !path_parts.contains_key(name) {
                let value = match file_name_parts.get(i) {
                    Some(v) => v,
                    None => {
                        return Err(Error::InvalidPath {
                            path: path.into(),
                            reason: "file name does not match expected structure".to_owned(),
                        })
                    }
                };
                path_parts.insert(name.clone(), Utf8PathBuf::from(value));
            }
        }

        Ok(Metadata::new(self, path_parts))
    }
}

/// Map of DRS Activity name to [`Structure`].
/// Used to validate paths against the known DRS structures and construct [`Metadata`] objects.
#[derive(Debug, Deserialize)]
pub struct Config {
    #[serde(flatten)]
    map: HashMap<Activity, Structure>,
}

impl Config {
    /// Creates a [`Metadata`] from the given path based on the available structures
    pub fn metadata_from_path<P: AsRef<Path>>(&self, path: &P) -> Result<Metadata, Error> {
        let path = path
            .as_ref()
            .canonicalize()
            .map_err(|_e| Error::CanonicalizePath {
                path: path.as_ref().to_owned(),
            })?;
        let path = Utf8PathBuf::try_from(path)?;

        let activity = match self.structure_from_path(&path) {
            Some(a) => a,
            None => {
                return Err(Error::InvalidPath {
                    path: path.into(),
                    reason: "no matching structure found".to_owned(),
                })
            }
        };
        let structure = &self.map[activity];
        debug!("found structure {} for path \n{}", activity.0, path);

        structure.metadata_from_path(&path)
    }

    /// Returns an iterator over the structures available in the config.
    pub fn structures(&self) -> impl Iterator<Item = &Structure> {
        self.map.values()
    }

    /// Constructs a new [`Config`] based on config file found at `filename`.
    pub fn from_toml<P: AsRef<Path>>(filename: P) -> Result<Self, Error> {
        let mut f = File::open(filename)?;
        let mut buffer = String::new();
        f.read_to_string(&mut buffer)?;

        Self::try_from(&buffer)
    }

    /// Searches for an [`Activity`] based on the given path
    pub fn structure_from_path<P: AsRef<Path>>(&self, path: P) -> Option<&Activity> {
        for (activity, structure) in self.map.iter() {
            if structure.verify_path(&path) {
                return Some(activity);
            }
        }

        None
    }
}

impl TryFrom<&str> for Config {
    type Error = Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        let mut s: Config = toml::from_str(value)?;
        for (_, mut structure) in s.map.iter_mut() {
            structure.root_dir = match structure.root_dir.canonicalize() {
                Ok(root_dir) => Utf8PathBuf::try_from(root_dir)?,
                Err(e) => {
                    warn!(
                        "error canonicalizing DRS root {}: {}",
                        structure.root_dir, e
                    );
                    continue;
                }
            };
        }
        Ok(s)
    }
}

impl TryFrom<&String> for Config {
    type Error = Error;

    fn try_from(value: &String) -> Result<Self, Self::Error> {
        let mut s: Config = toml::from_str(value)?;
        for (_, mut structure) in s.map.iter_mut() {
            structure.root_dir = match structure.root_dir.canonicalize() {
                Ok(root_dir) => Utf8PathBuf::try_from(root_dir)?,
                Err(e) => {
                    warn!(
                        "error canonicalizing DRS root {}: {}",
                        structure.root_dir, e
                    );
                    continue;
                }
            };
        }
        Ok(s)
    }
}

/// Filepath that conforms to the corresponding [`Structure`].
///
/// Paths must be UTF-8.
#[derive(Debug, PartialEq)]
pub struct Metadata<'a> {
    // in original, this is the string key rather than the actual structure but I think this can work here
    structure: &'a Structure,
    /// map of { part name => value } e.g. "institute" => "cpc"
    /// We assume that all file paths will be valid utf-8
    parts: HashMap<String, Utf8PathBuf>,
}

impl<'a> Metadata<'a> {
    fn new(structure: &'a Structure, parts: HashMap<String, Utf8PathBuf>) -> Self {
        Metadata { structure, parts }
    }

    /// Builds a fully qualified path from for this file.
    pub fn to_path(&self) -> Utf8PathBuf {
        let parts_order = &self.structure.parts_dir;
        let mut path = self.structure.root_dir.clone();
        for k in parts_order.iter() {
            path = path.join(&self.parts[k]);
        }

        path
    }

    /// Returns if this file is supposed to be versioned.
    pub fn versioned(&self) -> bool {
        self.structure.parts_versioned_dataset.is_some()
    }

    /// Returns the dataset identifier.
    ///
    /// Ref: DRS spec 3.4
    pub fn to_identifier(&self) -> String {
        // the original included a `versioned` param which would use `parts_versioned_dataset`
        // instead of `parts_dataset`. Every reference, with one exception, within the eval
        // system codebase explicitly set `versioned` to false and the only exception is a
        // function which passes through a `versioned` param that it accepts but it itself
        // does not appear to be called
        // I'm not sure this version stuff is necessary
        let parts = {
            if let Some(parts) = &self.structure.parts_versioned_dataset {
                parts
            } else {
                &self.structure.parts_dataset
            }
        };

        let mut results = Vec::new();
        for k in parts.iter() {
            if let Some(part) = self.parts.get(k) {
                results.push(part.as_str())
            } else if let Some(part) = self.structure.defaults.get(k) {
                results.push(part.as_str());
            }
        }

        results.join(".")
    }

    /// Returns the unversioned identifier i.e. ignores `parts_versioned_dataset` if
    /// it exists.
    pub fn to_unversioned_identifier(&self) -> String {
        let parts_keys = &self.structure.parts_dataset;
        let mut results = Vec::with_capacity(parts_keys.len());
        for k in parts_keys.iter() {
            if let Some(part) = self.parts.get(k) {
                results.push(part.as_str());
            } else if let Some(part) = self.structure.defaults.get(k) {
                results.push(part.as_str());
            }
            // should this error if the key isn't present in either?
        }

        results.join(".")
    }

    /// Returns a reference to the version part of the file data if present.
    pub fn version(&self) -> Option<&str> {
        self.parts.get("version").map(|v| v.as_str())
    }
}

impl<'a> Serialize for Metadata<'a> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut metadata = self.parts.clone();
        metadata.insert("file".to_owned(), self.to_path());
        if metadata.contains_key("version") {
            // replace is only on Strings but not paths so this awkward little dance is necessary.
            // Would be good to find a way to replace this.
            let version_to_remove = format!("/{}/", metadata["version"].as_str());
            let file_no_version = metadata["file"].as_str().replace(&version_to_remove, "/");

            metadata.insert(
                "file_no_version".to_owned(),
                Utf8PathBuf::from(file_no_version),
            );
        } else {
            metadata.insert("file_no_version".to_owned(), metadata["file"].to_owned());
        }

        metadata.insert("dataset".to_owned(), (&self.structure.data_type.0).into());

        // TODO: this doesn't add timestamp and creation_time which are added in solr_core.py
        // instead of with the rest of the metadata dictionary creation

        let mut map = serializer.serialize_map(Some(metadata.len()))?;
        for (k, v) in metadata {
            map.serialize_entry(&k, &v)?;
        }

        map.end()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    // rust-analyzer's single test debug uses a different directory from most other things which messes up
    // relative paths, so this helps us always keep paths relative to the root of the project rather than
    // relative to this file's directory
    const ROOT: &str = concat!(env!("CARGO_MANIFEST_DIR"), "/..");

    #[test]
    fn test_single_deserialization() {
        let config = load_single_config();

        assert!(config
            .map
            .contains_key(&Activity("observations".to_owned())));
        assert_eq!(
            config.map[&Activity("observations".to_owned())].data_type,
            Activity("observations".to_owned())
        );
    }

    #[test]
    fn test_single_file() {
        let config = load_single_config();
        let path = single_file();

        let res = config.metadata_from_path(&path);
        assert!(res.is_ok());
        let file = res.unwrap();

        let expected = Metadata {
            structure: &config.map[&Activity("observations".to_owned())],
            parts: [
                ("project", "observations"),
                ("product", "grid"),
                ("institute", "CPC"),
                ("model", "CPC"),
                ("experiment", "cmorph"),
                ("time_frequency", "30min"),
                ("realm", "atmos"),
                ("cmor_table", "30min"),
                ("ensemble", "r1i1p1"),
                ("version", "v20210618"),
                ("variable", "pr"),
                ("level", "cmorph"),
                ("time", "201609020000-201609020030"),
                (
                    "file_name",
                    "pr_30min_CPC_cmorph_r1i1p1_201609020000-201609020030.nc",
                ),
            ]
            .into_iter()
            .map(|(k, v)| (k.to_owned(), Utf8PathBuf::from(v)))
            .collect(),
        };

        assert_eq!(expected, file);
    }

    #[test]
    fn test_identifier() {
        let config = load_single_config();
        let path = single_file();

        let res = config.metadata_from_path(&path);
        assert!(res.is_ok());
        let file = res.unwrap();

        let expected = "observations.grid.CPC.CPC.cmorph.30min.atmos.30min.r1i1p1.v20210618.pr";
        assert_eq!(expected, file.to_identifier());
    }

    #[test]
    fn test_unversioned_identifier() {
        let config = load_single_config();
        let path = single_file();

        let res = config.metadata_from_path(&path);
        assert!(res.is_ok());
        let file = res.unwrap();

        let expected = "observations.grid.CPC.CPC.cmorph.30min.atmos.30min.r1i1p1.pr";
        assert_eq!(expected, file.to_unversioned_identifier());
    }

    #[test]
    fn test_nonexistant_file() {
        let config = load_single_config();
        let path = "this/file/doesnt/exist.txt";

        let result = config.metadata_from_path(&path);
        assert!(
            match result.unwrap_err() {
                Error::CanonicalizePath { path: _ } => true,
                _ => false,
            },
            "incorrect error type"
        );
    }

    #[test]
    fn test_incorrect_file_name_parts() {
        // this isn't necessarily correct behavior but it is a case to be aware of

        let mut config = load_single_config();
        let path = single_file();

        // append another segment to the structure's file name parts so the `file_from_path` fails
        let observations = config
            .map
            .get_mut(&Activity("observations".to_owned()))
            .unwrap();
        observations.parts_file_name.push("foo".to_owned());

        let res = config.metadata_from_path(&path);
        assert!(res.is_err());
        assert!(match res.unwrap_err() {
            Error::InvalidPath { path: _, reason: _ } => true,
            _ => false,
        });
    }

    fn load_single_config() -> Config {
        let data = single_config();
        Config::try_from(&data).unwrap()
    }

    fn single_config() -> String {
        format!(
            r#"
            [observations]
                root_dir = "{ROOT}/.docker/data"
                parts_dir = [
                    "project", "product", "institute", "model", "experiment", "time_frequency", "realm",
                    "cmor_table", "ensemble", "version", "variable", "file_name"
                ]
                parts_dataset = [
                    "project", "product", "institute", "model", "experiment", "time_frequency", "realm",
                    "cmor_table", "ensemble", "", "variable"
                ]
                parts_versioned_dataset = [
                    "project", "product", "institute", "model", "experiment", "time_frequency", "realm",
                    "cmor_table", "ensemble", "version", "variable"
                ]
                parts_file_name = ["variable", "time_frequency", "experiment", "level", "version", "time"]
                parts_time = "start_time-end_time"
                data_type = "observations"

                [observations.defaults]
                project = "observations"
            "#
        )
    }

    fn single_file() -> Utf8PathBuf {
        Utf8PathBuf::from(
            format!(
                concat!(
                    "{root}/.docker/data/observations/grid/CPC/CPC/cmorph/30min/atmos/30min/r1i1p1/v20210618/pr",
                    "/pr_30min_CPC_cmorph_r1i1p1_201609020000-201609020030.nc"
                ),
            root=ROOT,
            )
        )
    }
}
